@@ -4,8 +4,9 @@
 #   Tree visitor and transform framework
 #
 
-from __future__ import absolute_import
+from __future__ import absolute_import, print_function
 
+import sys
 import inspect
 
 from . import TypeSlots
@@ -14,10 +15,19 @@ from . import Nodes
 from . import ExprNodes
 from . import Errors
 from . import DebugFlags
+from . import Future
 
 import cython
 
 
+cython.declare(_PRINTABLE=tuple)
+
+if sys.version_info[0] >= 3:
+    _PRINTABLE = (bytes, str, int, float)
+else:
+    _PRINTABLE = (str, unicode, long, int, float)
+
+
 class TreeVisitor(object):
     """
     Base class for writing visitors for a Cython tree, contains utilities for
@@ -47,9 +57,9 @@ class TreeVisitor(object):
     >>> tree = SampleNode(0, SampleNode(1), [SampleNode(2), SampleNode(3)])
     >>> class MyVisitor(TreeVisitor):
     ...     def visit_SampleNode(self, node):
-    ...         print "in", node.value, self.access_path
+    ...         print("in %s %s" % (node.value, self.access_path))
     ...         self.visitchildren(node)
-    ...         print "out", node.value
+    ...         print("out %s" % node.value)
     ...
     >>> MyVisitor().visit(tree)
     in 0 []
@@ -93,7 +103,7 @@ class TreeVisitor(object):
                 continue
             elif isinstance(value, list):
                 value = u'[...]/%d' % len(value)
-            elif not isinstance(value, (str, unicode, long, int, float)):
+            elif not isinstance(value, _PRINTABLE):
                 continue
             else:
                 value = repr(value)
@@ -152,11 +162,11 @@ class TreeVisitor(object):
             handler_method = getattr(self, pattern % mro_cls.__name__, None)
             if handler_method is not None:
                 return handler_method
-        print type(self), cls
+        print(type(self), cls)
         if self.access_path:
-            print self.access_path
-            print self.access_path[-1][0].pos
-            print self.access_path[-1][0].__dict__
+            print(self.access_path)
+            print(self.access_path[-1][0].pos)
+            print(self.access_path[-1][0].__dict__)
         raise RuntimeError("Visitor %r does not accept object: %s" % (self, obj))
 
     def visit(self, obj):
@@ -175,7 +185,7 @@ class TreeVisitor(object):
             raise
         except Errors.AbortError:
             raise
-        except Exception, e:
+        except Exception as e:
             if DebugFlags.debug_no_exception_intercept:
                 raise
             self._raise_compiler_error(obj, e)
@@ -239,7 +249,7 @@ class VisitorTransform(TreeVisitor):
     """
     def visitchildren(self, parent, attrs=None):
         result = self._visitchildren(parent, attrs)
-        for attr, newnode in result.iteritems():
+        for attr, newnode in result.items():
             if type(newnode) is not list:
                 setattr(parent, attr, newnode)
             else:
@@ -434,7 +444,7 @@ find_special_method_for_binary_operator = {
     '>':  '__gt__',
     '+':  '__add__',
     '&':  '__and__',
-    '/':  '__truediv__',
+    '/':  '__div__',
     '//': '__floordiv__',
     '<<': '__lshift__',
     '%':  '__mod__',
@@ -515,6 +525,9 @@ class MethodDispatcherTransform(EnvTransform):
             operand1, operand2 = node.operand1, node.operand2
             if special_method_name == '__contains__':
                 operand1, operand2 = operand2, operand1
+            elif special_method_name == '__div__':
+                if Future.division in self.current_env().global_scope().context.future_directives:
+                    special_method_name = '__truediv__'
             obj_type = operand1.type
             if obj_type.is_builtin_type:
                 type_name = obj_type.name
@@ -590,9 +603,19 @@ class MethodDispatcherTransform(EnvTransform):
                 return function_handler(node, function, arg_list, kwargs)
             else:
                 return function_handler(node, function, arg_list)
-        elif function.is_attribute and function.type.is_pyobject:
+        elif function.is_attribute:
             attr_name = function.attribute
-            self_arg = function.obj
+            if function.type.is_pyobject:
+                self_arg = function.obj
+            elif node.self and function.entry:
+                entry = function.entry.as_variable
+                if not entry or not entry.is_builtin:
+                    return node
+                # C implementation of a Python builtin method - see if we find further matches
+                self_arg = node.self
+                arg_list = arg_list[1:]  # drop CloneNode of self argument
+            else:
+                return node
             obj_type = self_arg.type
             is_unbound_method = False
             if obj_type.is_builtin_type:
@@ -630,11 +653,12 @@ class MethodDispatcherTransform(EnvTransform):
         if self_arg is not None:
             arg_list = [self_arg] + list(arg_list)
         if kwargs:
-            return method_handler(
+            result = method_handler(
                 node, function, arg_list, is_unbound_method, kwargs)
         else:
-            return method_handler(
+            result = method_handler(
                 node, function, arg_list, is_unbound_method)
+        return result
 
     def _handle_function(self, node, function_name, function, arg_list, kwargs):
         """Fallback handler"""
@@ -705,12 +729,17 @@ class PrintTree(TreeVisitor):
     """Prints a representation of the tree to standard output.
     Subclass and override repr_of to provide more information
     about nodes. """
-    def __init__(self):
+    def __init__(self, start=None, end=None):
         TreeVisitor.__init__(self)
         self._indent = ""
+        if start is not None or end is not None:
+            self._line_range = (start or 0, end or 2**30)
+        else:
+            self._line_range = None
 
     def indent(self):
         self._indent += "  "
+
     def unindent(self):
         self._indent = self._indent[:-2]
 
@@ -724,15 +753,17 @@ class PrintTree(TreeVisitor):
     # under the parent-node, not displaying the list itself in
     # the hierarchy.
     def visit_Node(self, node):
-        if len(self.access_path) == 0:
-            name = "(root)"
-        else:
-            parent, attr, idx = self.access_path[-1]
-            if idx is not None:
-                name = "%s[%d]" % (attr, idx)
+        line = node.pos[1]
+        if self._line_range is None or self._line_range[0] <= line <= self._line_range[1]:
+            if len(self.access_path) == 0:
+                name = "(root)"
             else:
-                name = attr
-        print("%s- %s: %s" % (self._indent, name, self.repr_of(node)))
+                parent, attr, idx = self.access_path[-1]
+                if idx is not None:
+                    name = "%s[%d]" % (attr, idx)
+                else:
+                    name = attr
+            print("%s- %s: %s" % (self._indent, name, self.repr_of(node)))
         self.indent()
         self.visitchildren(node)
         self.unindent()
